#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

import os
import platform
import re
import shutil
import subprocess
import sys
from asyncio import create_subprocess_exec
from asyncio.subprocess import PIPE, Process
from dataclasses import fields
from json import loads
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import minizinc

from .error import ConfigurationError, parse_error
from .json import decode_json_stream
from .solver import Solver

#: MiniZinc version required by the python package
CLI_REQUIRED_VERSION = (2, 5, 4)
#: Default locations on MacOS where the MiniZinc packaged release would be installed
MAC_LOCATIONS = [
    str(Path("/Applications/MiniZincIDE.app/Contents/Resources")),
    str(Path("~/Applications/MiniZincIDE.app/Contents/Resources").expanduser()),
]
#: Default locations on Windows where the MiniZinc packaged release would be installed
WIN_LOCATIONS = [
    str(Path("c:/Program Files/MiniZinc")),
    str(Path("c:/Program Files/MiniZinc IDE (bundled)")),
    str(Path("c:/Program Files (x86)/MiniZinc")),
    str(Path("c:/Program Files (x86)/MiniZinc IDE (bundled)")),
]


class Driver:
    """Driver that interfaces with MiniZinc through the command line interface.

    The command line driver will interact with MiniZinc and its solvers through
    the use of a ``minizinc`` executable. Driving MiniZinc using its executable
    is non-incremental and can often trigger full recompilation and might
    restart the solver from the beginning when changes are made to the instance.

    Raises:
        ConfigurationError: If an the driver version is found to be incompatible with
            MiniZinc Python

    Attributes:
        _executable (Path): The path to the executable used to access the MiniZinc
    """

    _executable: Path
    _solver_cache: Optional[Dict[str, List[Solver]]] = None
    _version: Optional[Tuple[int, ...]] = None

    def __init__(self, executable: Path):
        self._executable = executable
        if not self._executable.exists():
            raise ConfigurationError(
                f"No MiniZinc executable was found at '{self._executable}'."
            )

        if self.parsed_version < CLI_REQUIRED_VERSION:
            raise ConfigurationError(
                f"The MiniZinc driver found at '{self._executable}' has "
                f"version {self.parsed_version}. The minimal required version is "
                f"{CLI_REQUIRED_VERSION}."
            )

    def make_default(self) -> None:
        """Method to override the current default MiniZinc Python driver with the
        current driver.
        """
        minizinc.default_driver = self

    @property
    def executable(self) -> Path:
        """Reports the Path of the MiniZinc executable used by the Driver object

        Returns:
            Path: location of the MiniZinc executable
        """
        return self._executable

    @property
    def minizinc_version(self) -> str:
        """Reports the version text of the MiniZinc Driver

        Report the full version text of MiniZinc as reported by the driver,
        including the driver name, the semantic version, the build reference,
        and its authors.

        Returns:
            str: the version of as reported by the MiniZinc driver
        """
        # Note: cannot use "_run" as it already required the parsed version
        return subprocess.run(
            [str(self._executable), "--version"],
            stdin=None,
            stdout=PIPE,
            stderr=PIPE,
        ).stdout.decode()

    @property
    def parsed_version(self) -> Tuple[int, ...]:
        """Reports the version of the MiniZinc Driver

        Report the parsed version of the MiniZinc Driver as a tuple of integers.
        The tuple is ordered: major, minor, patch.

        Returns:
            Tuple[int, ...]: the parsd version reported by the MiniZinc driver
        """
        if self._version is None:
            match = re.search(r"version (\d+)\.(\d+)\.(\d+)", self.minizinc_version)
            assert match
            self._version = tuple([int(i) for i in match.groups()])
        return self._version

    def available_solvers(self, refresh=False):
        """Returns a list of available solvers

        This method returns the list of solvers available to the Driver object
        according to the current environment. Note that the list of solvers might
        be cached for future usage. The refresh argument can be used to ignore
        the current cache.

        Args:
            refresh (bool): When set to true, the Driver will rediscover the
                available solvers from the current environment.

        Returns:
            Dict[str, List[Solver]]: A dictionary that maps solver tags to MiniZinc
                solver configurations that can be used with the Driver object.
        """
        if not refresh and self._solver_cache is not None:
            return self._solver_cache

        # Find all available solvers
        output = self._run(["--solvers-json"])
        solvers = loads(output.stdout)

        # Construct Solver objects
        self._solver_cache = {}
        allowed_fields = {f.name for f in fields(Solver)}
        for s in solvers:
            obj = Solver(
                **{key: value for (key, value) in s.items() if key in allowed_fields}
            )
            if obj.version == "<unknown version>":
                obj._identifier = obj.id
            else:
                obj._identifier = obj.id + "@" + obj.version

            names = s.get("tags", [])
            names.extend([s["id"], s["id"].split(".")[-1]])
            for name in names:
                self._solver_cache.setdefault(name, []).append(obj)

        return self._solver_cache

    def _run(
        self,
        args: List[Any],
        solver: Optional[Solver] = None,
    ):
        """Start a driver process with given arguments

        Args:
            args (List[str]): direct arguments to the driver
            solver (Union[str, Path, None]): Solver configuration string
                guaranteed by the user to be valid until the process has ended.
        """
        # TODO: Add documentation
        windows_spawn_options: Dict[str, Any] = {}
        if sys.platform == "win32":
            # On Windows, MiniZinc terminates its subprocesses by generating a
            # Ctrl+C event for its own console using GenerateConsoleCtrlEvent.
            # Therefore, we must spawn it in its own console to avoid receiving
            # that Ctrl+C ourselves.
            #
            # On POSIX systems, MiniZinc terminates its subprocesses by sending
            # SIGTERM to the solver's process group, so this workaround is not
            # necessary as we won't receive that signal.
            windows_spawn_options = {
                "startupinfo": subprocess.STARTUPINFO(
                    dwFlags=subprocess.STARTF_USESHOWWINDOW,
                    wShowWindow=subprocess.SW_HIDE,
                ),
                "creationflags": subprocess.CREATE_NEW_CONSOLE,
            }

        # TODO: Always add --json-stream once 2.6.0 is minimum requirement
        if self.parsed_version >= (2, 6, 0):
            args.append("--json-stream")

        if solver is None:
            cmd = [str(self._executable), "--allow-multiple-assignments"] + [
                str(arg) for arg in args
            ]
            minizinc.logger.debug(f"CLIDriver:run -> command: \"{' '.join(cmd)}\"")
            output = subprocess.run(
                cmd,
                stdin=None,
                stdout=PIPE,
                stderr=PIPE,
                **windows_spawn_options,
            )
        else:
            with solver.configuration() as conf:
                cmd = [
                    str(self._executable),
                    "--solver",
                    conf,
                    "--allow-multiple-assignments",
                ] + [str(arg) for arg in args]
                minizinc.logger.debug(f"CLIDriver:run -> command: \"{' '.join(cmd)}\"")
                output = subprocess.run(
                    cmd,
                    stdin=None,
                    stdout=PIPE,
                    stderr=PIPE,
                    **windows_spawn_options,
                )
        if output.returncode != 0:
            if self.parsed_version >= (2, 6, 0):
                # Error will (usually) be raised in json stream
                for _ in decode_json_stream(output.stdout):
                    pass
            raise parse_error(output.stderr)
        return output

    async def _create_process(
        self, args: List[str], solver: Optional[str] = None
    ) -> Process:
        """Start an asynchronous driver process with given arguments

        Args:
            args (List[str]): direct arguments to the driver
            solver (Union[str, Path, None]): Solver configuration string
                guaranteed by the user to be valid until the process has ended.
        """

        windows_spawn_options: Dict[str, Any] = {}
        if sys.platform == "win32":
            # See corresponding comment in run()
            windows_spawn_options = {
                "startupinfo": subprocess.STARTUPINFO(
                    dwFlags=subprocess.STARTF_USESHOWWINDOW,
                    wShowWindow=subprocess.SW_HIDE,
                ),
                "creationflags": subprocess.CREATE_NEW_CONSOLE,
            }

        # TODO: Always add --json-stream once 2.6.0 is minimum requirement
        if self.parsed_version >= (2, 6, 0):
            args.append("--json-stream")

        if solver is None:
            minizinc.logger.debug(
                f"CLIDriver:create_process -> program: {str(self._executable)} "
                f'args: "--allow-multiple-assignments '
                f"{' '.join(str(arg) for arg in args)}\""
            )
            proc = await create_subprocess_exec(
                str(self._executable),
                "--allow-multiple-assignments",
                *[str(arg) for arg in args],
                stdin=None,
                stdout=PIPE,
                stderr=PIPE,
                **windows_spawn_options,
            )
        else:
            minizinc.logger.debug(
                f"CLIDriver:create_process -> program: {str(self._executable)} "
                f'args: "--solver {solver} --allow-multiple-assignments '
                f"{' '.join(str(arg) for arg in args)}\""
            )
            proc = await create_subprocess_exec(
                str(self._executable),
                "--solver",
                solver,
                "--allow-multiple-assignments",
                *[str(arg) for arg in args],
                stdin=None,
                stdout=PIPE,
                stderr=PIPE,
                **windows_spawn_options,
            )
        return proc

    @classmethod
    def find(
        cls, path: Optional[List[str]] = None, name: str = "minizinc"
    ) -> Optional["Driver"]:
        """Finds MiniZinc Driver on default or specified path.

        Find driver will look for the MiniZinc executable to create a Driver for
        MiniZinc Python. If no path is specified, then the paths given by the
        environment variables appended by MiniZinc's default locations will be tried.

        Args:
            path: List of locations to search.
            name: Name of the executable.

        Returns:
            Optional[Driver]: Returns a Driver object when found or None.
        """
        if path is None:
            path = os.environ.get("PATH", "").split(os.pathsep)
            # Add default MiniZinc locations to the path
            if platform.system() == "Darwin":
                path.extend(MAC_LOCATIONS)
            elif platform.system() == "Windows":
                path.extend(WIN_LOCATIONS)

        # Try to locate the MiniZinc executable
        executable = shutil.which(name, path=os.pathsep.join(path))
        if executable is not None:
            return cls(Path(executable))
        return None



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

import re
from dataclasses import dataclass
from datetime import timedelta
from enum import Enum, auto
from json import loads
from typing import Any, Dict, List, Optional, Tuple, Type, Union

from .json import MZNJSONDecoder
from .model import Method

StatisticsType = Union[float, int, str, timedelta]

StdStatisticTypes = {
    # Number of search nodes
    "nodes": int,
    # Number of leaf nodes that were failed
    "failures": int,
    # Number of times the solver restarted the search
    "restarts": int,
    # Number of variables
    "variables": int,
    # Number of integer variables created by the solver
    "intVariables": int,
    # Number of Boolean variables created by the solver
    "boolVariables": int,
    # Number of floating point variables created by the solver
    "floatVariables": int,
    # Number of set variables created by the solver
    "setVariables": int,
    # Number of propagators created by the solver
    "propagators": int,
    # Number of propagator invocations
    "propagations": int,
    # Peak depth of search tree
    "peakDepth": int,
    # Number of nogoods created
    "nogoods": int,
    # Number of backjumps
    "backjumps": int,
    # Peak memory (in Mbytes)
    "peakMem": float,
    # Initialisation time
    "initTime": timedelta,
    # Solving time
    "solveTime": timedelta,
    # Flattening time
    "flatTime": timedelta,
    # Number of paths generated
    "paths": int,
    # Number of Boolean variables in the flat model
    "flatBoolVars": int,
    # Number of floating point variables in the flat model
    "flatFloatVars": int,
    # Number of integer variables in the flat model
    "flatIntVars": int,
    # Number of set variables in the flat model
    "flatSetVars": int,
    # Number of Boolean constraints in the flat model
    "flatBoolConstraints": int,
    # Number of floating point constraints in the flat model
    "flatFloatConstraints": int,
    # Number of integer constraints in the flat model
    "flatIntConstraints": int,
    # Number of set constraints in the flat model
    "flatSetConstraints": int,
    # Optimisation method in the Flat Model
    "method": str,
    # Number of reified constraints evaluated during flattening
    "evaluatedReifiedConstraints": int,
    # Number of half-reified constraints evaluated during flattening
    "evaluatedHalfReifiedConstraints": int,
    # Number of implications removed through chain compression
    "eliminatedImplications": int,
    # Number of linear constraints removed through chain compression
    "eliminatedLinearConstraints": int,
}


def set_stat(stats: Dict[str, StatisticsType], name: str, value: str):
    """Set statistical value in the result object.

    Set solving statiscal value within the result object. This value is
    converted from string to the appropriate type as suggested by the
    statistical documentation in MiniZinc. Timing statistics, expressed through
    floating point numbers in MiniZinc, will be converted to timedelta objects.

    Args:
        stats: The dictionary to be extended with the new statistical value
        name: The name under which the statistical value will be stored.
        value: The value for the statistical value to be converted and stored.

    """
    value = value.strip('"')
    tt = StdStatisticTypes.get(name, None)
    if tt is None and ("time" in name or "Time" in name):
        tt = timedelta
    try:
        if tt is timedelta:
            time_us = int(float(value) * 1000000)
            stats[name] = timedelta(microseconds=time_us)
        elif tt is not None:
            stats[name] = tt(value)
        else:
            try:
                stats[name] = int(value)
            except ValueError:
                stats[name] = float(value)
    except ValueError:
        stats[name] = value


class Status(Enum):
    """Enumeration to represent the status of the solving process.

    Attributes:
        ERROR: An error occurred during the solving process.
        UNKNOWN: No solutions have been found and search has terminated without
            exploring the whole search space.
        UNBOUNDED: The objective of the optimisation problem is unbounded.
        UNSATISFIABLE: No solutions have been found and the whole search space
            was explored.
        SATISFIED: A solution was found, but possibly not the whole search
            space was explored.
        ALL_SOLUTIONS: All solutions in the search space have been found.
        OPTIMAL_SOLUTION: A solution has been found that is optimal according
            to the objective.

    """

    ERROR = auto()
    UNKNOWN = auto()
    UNBOUNDED = auto()
    UNSATISFIABLE = auto()
    SATISFIED = auto()
    ALL_SOLUTIONS = auto()
    OPTIMAL_SOLUTION = auto()

    @classmethod
    def from_output(cls, output: bytes, method: Method):
        """Determines the solving status from the output of a MiniZinc process

        Determines the solving status according to the standard status output
        strings defined by MiniZinc. The output of the MiniZinc will be scanned
        in a defined order to best determine the status of the solving process.
        UNKNOWN will be returned if no status can be determined.

        Args:
            output (bytes): the standard output of a MiniZinc process.
            method (Method): the objective method used in the optimisation
                problem.

        Returns:
            Optional[Status]: Status that could be determined from the output.

        """
        s = None
        if b"=====ERROR=====" in output:
            s = cls.ERROR
        elif b"=====UNKNOWN=====" in output:
            s = cls.UNKNOWN
        elif b"=====UNSATISFIABLE=====" in output:
            s = cls.UNSATISFIABLE
        elif (
            b"=====UNSATorUNBOUNDED=====" in output or b"=====UNBOUNDED=====" in output
        ):
            s = cls.UNBOUNDED
        elif method is Method.SATISFY:
            if b"==========" in output:
                s = cls.ALL_SOLUTIONS
            elif b"----------" in output:
                s = cls.SATISFIED
        else:
            if b"==========" in output:
                s = cls.OPTIMAL_SOLUTION
            elif b"----------" in output:
                s = cls.SATISFIED
        return s

    @classmethod
    def from_str(cls, status: str):
        s = None
        if status == "ERROR":
            s = cls.ERROR
        elif status == "UNKNOWN":
            s = cls.UNKNOWN
        elif status == "UNBOUNDED" or status == "UNSAT_OR_UNBOUNDED":
            s = cls.UNBOUNDED
        elif status == "UNSATISFIABLE":
            s = cls.UNSATISFIABLE
        elif status == "SATISFIED":
            s = cls.SATISFIED
        elif status == "ALL_SOLUTIONS":
            s = cls.ALL_SOLUTIONS
        elif status == "OPTIMAL_SOLUTION":
            s = cls.OPTIMAL_SOLUTION
        return s

    def __str__(self):
        return self.name

    def has_solution(self) -> bool:
        """Returns true if the status suggest that a solution has been found."""
        if self in [self.SATISFIED, self.ALL_SOLUTIONS, self.OPTIMAL_SOLUTION]:
            return True
        return False


@dataclass
class Result:
    """Representation of a MiniZinc solution in Python

    Attributes:
        status (Status): The solving status of the MiniZinc instance
        solution (Any): Variable assignments
            made to form the solution
        statistics (Dict[str, Union[float, int, timedelta]]): Statistical
            information generated during the search for the Solution

    """

    status: Status
    solution: Any
    statistics: Dict[str, Union[float, int, timedelta]]

    @property
    def objective(self) -> Optional[Union[int, float]]:
        """Returns objective of the solution

        Returns the objective of the solution when possible. If no solutions
        have been found or the problem did not have an objective, then None is
        returned instead.

        Returns:
            Optional[Union[int, float]]: best objective found or None

        """
        if self.solution is not None:
            if isinstance(self.solution, list):
                return getattr(self.solution[-1], "objective", None)
            else:
                return getattr(self.solution, "objective", None)
        else:
            return None

    def __getitem__(self, key):
        """Retrieves solution or a member of a solution.

        Overrides the default implementation of item access (obj[key]) to
        retrieve a solution object or member of a solution from the result
        object.

        - If the Result object does not contain any solutions, then a
          KeyError will always be raised.
        - If the Result object contains a single solutions, then the names of a
          variable can be used in this method to retrieve its value in the
          solution.
        - If the Result object contains multiple solutions, then a single
          integer can be used to retrieve the solution object or a tuple of an
          integer and the name of a variable can be used to retrieve the value
          of that variable in the numbered solution object.

        Args:
            key: solution number or name of the solution member.

        Returns:
            Solution object or the value of the member in the solution.

        Raises:
            KeyError: No solution was found, solution number is out of range,
                or no solution member with this name exists.

        """
        try:
            if self.solution is not None:
                if isinstance(self.solution, list):
                    if isinstance(key, tuple):
                        return getattr(self.solution.__getitem__(key[0]), key[1])
                    else:
                        return self.solution.__getitem__(key)
                else:
                    return getattr(self.solution, key)
            else:
                raise KeyError
        except AttributeError:
            raise KeyError

    def __len__(self):
        """Returns the number of solutions included in the Result object

        Returns:
            int: number of solution that can be accessed

        """
        if self.solution is None:
            return 0
        elif isinstance(self.solution, list):
            return len(self.solution)
        else:
            return 1

    def __str__(self):
        return str(self.solution)


def parse_solution(
    raw: bytes,
    output_type: Type,
    enum_map: Optional[Dict[str, Enum]] = None,
    renames: Optional[List[Tuple[str, str]]] = None,
) -> Tuple[Optional[Any], Dict]:
    """Parses a solution from the output of a MiniZinc process.

    Parses the MiniZinc output between solution separators. The solution is
    expected to be formatted according to the MiniZinc JSON standards.
    Statistical information can be included in the output in accordance with
    the MiniZinc statistics standard. Statistics will be parsed and retyped
    according to the types given in StatisticTypes. Statistics will be parsed
    even if no solution is found.

    Args:
        raw (bytes): The output on stdout for one solution of the process
            solving the MiniZinc instance.
        output_type (Type): The type used for every solution
        enum_map (Optional[Dict[str, Enum]]): A map to map enumeration identifiers to
            the internal values used in Python
        ranames (Optional[List[Tuple[str, str]]]): A list of keys to be renamed from
            the raw solution to the solution object input

    Returns:
        Tuple[Optional[Dict], Dict]: A tuple containing the parsed solution
            assignments and the parsed statistics.

    """
    if enum_map is None:
        enum_map = {}
    if renames is None:
        renames = []
    # Parse statistics
    statistics: Dict[str, StatisticsType] = {}
    matches = re.findall(rb"%%%mzn-stat:? (\w*)=([^\r\n]*)", raw)
    for m in matches:
        set_stat(statistics, m[0].decode(), m[1].decode())
    match = re.search(rb"% time elapsed: (\d+.\d+) s", raw)
    if match:
        time_us = int(float(match[1]) * 1000000)
        statistics["time"] = timedelta(microseconds=time_us)

    # Parse solution
    solution = None
    raw = re.sub(
        rb"^-{10}|={5}(ERROR|UNKNOWN|UNSATISFIABLE|UNSATorUNBOUNDED|UNBOUNDED|)?={5}",
        b"",
        raw,
        flags=re.MULTILINE,
    )
    raw = re.sub(rb"^\w*%.*\n?", b"", raw, flags=re.MULTILINE)
    if b"{" in raw:
        tmp = loads(raw, enum_map=enum_map, cls=MZNJSONDecoder)
        if "_objective" in tmp:
            tmp["objective"] = tmp.pop("_objective")
        if "_output" in tmp:
            tmp["_output_item"] = tmp.pop("_output")
        for before, after in renames:
            tmp[after] = tmp.pop(before)

        solution = output_type(**tmp)

    return solution, statistics



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

import json
import threading
import warnings
from enum import Enum, EnumMeta
from pathlib import Path
from typing import Any, Dict, List, Optional, Type, Union

ParPath = Union[Path, str]


class Method(Enum):
    """Enumeration that represents of a solving method.

    Attributes:
        SATISFY: Represents a satisfaction problem.
        MINIMIZE: Represents a minimization problem.
        MAXIMIZE: Represents a maximization problem.
    """

    SATISFY = 1
    MINIMIZE = 2
    MAXIMIZE = 3

    @classmethod
    def from_string(cls, s: str):
        """Get Method represented by the string s.

        Args:
            s: String expected to contain either "sat", "min", or "max".

        Returns:
            Method: Method represented by s
        """
        if s == "sat":
            return cls.SATISFY
        elif s == "min":
            return cls.MINIMIZE
        elif s == "max":
            return cls.MAXIMIZE
        else:
            raise ValueError(
                f"Unknown Method {s}, valid options are 'sat', 'min', or 'max'"
            )


class UnknownExpression(str):
    pass


class Model:
    """The representation of a MiniZinc model in Python

    Attributes:
        output_type (Type): the type used to store the solution values created
            in the process of solving the Instance. This attribute is
            particularly helpful when comparing the results of multiple
            instances together. The type must support initialisation with the
            assignments returned by MiniZinc. These assignments currently
            always include "__output_item" and include "objective" if the
            instance is not a satisfaction problem.

    Raises:
        MiniZincError: when an error occurs during the parsing or
            type checking of the model object.
    """

    output_type: Optional[Type] = None

    _code_fragments: List[str]
    _data: Dict[str, Any]
    _enum_map: Dict[str, Enum]
    _includes: List[Path]
    _lock: threading.Lock
    _checker: bool = False

    def __init__(self, files: Optional[Union[ParPath, List[ParPath]]] = None):
        self._data = {}
        self._includes = []
        self._code_fragments = []
        self._enum_map = {}
        self._lock = threading.Lock()
        if isinstance(files, Path) or isinstance(files, str):
            self._add_file(files)
        elif files is not None:
            for file in files:
                self._add_file(file)

    def __setitem__(self, key: str, value: Any):
        """Set parameter of Model.

        This method overrides the default implementation of item access
        (``obj[key] = value``) for models. Item access on a Model can be used to
        set parameters of the Model.

        Args:
            key (str): Identifier of the parameter.
            value (Any): Value to be assigned to the parameter.
        """
        with self._lock:
            if self._data.get(key, None) is None:
                if isinstance(value, EnumMeta):
                    self._register_enum_values(value)
                self._data.__setitem__(key, value)
            else:
                if self._data[key] != value:
                    # TODO: Fix the error type and document
                    raise AssertionError(
                        f"The parameter '{key}' cannot be assigned multiple values. "
                        f"If you are changing the model, consider using the branch "
                        f"method before assigning the parameter."
                    )

    def _register_enum_values(self, t: EnumMeta):
        for name in t.__members__:
            if name in self._enum_map:
                # TODO: Fix the error type and document
                raise AssertionError(
                    f"Identifier '{name}' is used in multiple enumerated types"
                    f"within the same model. Identifiers in enumerated types "
                    f"have to be unique."
                )
            self._enum_map[name] = t.__members__[name]

    def __getitem__(self, key: str) -> Any:
        """Get parameter of Model.

        This method overrides the default implementation of item access
        (``obj[key]``) for models. Item access on a Model can be used to get
        parameters of the Model.

        Args:
            key (str): Identifier of the parameter.

        Returns:
            The value assigned to the parameter.

        Raises:
            KeyError: The parameter you are trying to access is not known.

        """
        return self._data.__getitem__(key)

    def add_file(self, file: ParPath, parse_data: bool = False) -> None:
        """Adds a MiniZinc file (``.mzn``, ``.dzn``, or ``.json``) to the Model.

        Args:
            file (Union[Path, str]): Path to the file to be added
            parse_data (bool): Signal if the data should be parsed for usage
                within Python. This option is ignored if the extra `dzn` is
                not enabled.
        Raises:
            MiniZincError: when an error occurs during the parsing or
                type checking of the model object.
        """
        return self._add_file(file, parse_data)

    def _add_file(self, file: ParPath, parse_data: bool = False) -> None:
        if not isinstance(file, Path):
            file = Path(file)
        assert file.exists()
        if not parse_data:
            with self._lock:
                self._includes.append(file)
            return
        if file.suffix == ".json":
            data = json.load(file.open())
            for k, v in data.items():
                self.__setitem__(k, v)
        elif file.suffix == ".dzn" and parse_data:
            try:
                from lark.exceptions import LarkError

                from .dzn import parse_dzn

                try:
                    data = parse_dzn(file)
                    for k, v in data.items():
                        self.__setitem__(k, v)
                except LarkError:
                    warnings.warn(
                        f"Could not parse {file}. Parameters included within this file "
                        f"are not available in Python"
                    )
                    with self._lock:
                        self._includes.append(file)
            except ImportError:
                with self._lock:
                    self._includes.append(file)
        elif file.suffix not in [".dzn", ".mzn", ".mzc"]:
            raise NameError("Unknown file suffix %s", file.suffix)
        else:
            with self._lock:
                if ".mzc" in file.suffixes:
                    self._checker = True
                self._includes.append(file)

    def add_string(self, code: str) -> None:
        """Adds a string of MiniZinc code to the Model.

        Args:
            code (str): A string contain MiniZinc code
        Raises:
            MiniZincError: when an error occurs during the parsing or
                type checking of the model object.
        """
        with self._lock:
            self._code_fragments.append(code)

    def __copy__(self):
        copy = self.__class__()
        copy._includes = self._includes[:]
        copy._code_fragments = self._code_fragments[:]
        copy._data = dict.copy(self._data)
        return copy



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

from dataclasses import dataclass
from typing import Any


@dataclass(frozen=True)
class AnonEnum:
    """Representation of anonymous enumeration values in MiniZinc"""

    enumName: str
    value: int

    def __str__(self):
        return f"to_enum({self.enumName},{self.value})"


@dataclass(frozen=True)
class ConstrEnum:
    """Representation of constructor function enumerated values in MiniZinc"""

    constructor: str
    argument: Any

    def __str__(self):
        return f"{self.constructor}({self.argument})"



#  This Source Code Form is subject to the terms of the Mozilla Public
#  License, v. 2.0. If a copy of the MPL was not distributed with this
#  file, You can obtain one at http://mozilla.org/MPL/2.0/.

from pathlib import Path
from typing import Union

from lark import Lark, Transformer

from minizinc.model import UnknownExpression

dzn_grammar = r"""
    items: [item (";" item)* ";"?]
    item: ident "=" value | ident "=" unknown
    ident: /([A-Za-z][A-Za-z0-9_]*)|(\'[^\']*\')/
    value: array
         | array2d
         | set
         | int
         | float
         | string
         | "true"       -> true
         | "false"      -> false
    list: [value ("," value)* ","?]
    array: "[" list "]"
    array2d: "[" "|" [ list ("|" list)*] "|" "]"
    set: "{" list "}"
       | int ".." int

    int: /-?((0o[0-7]+)|(0x[0-9A-Fa-f]+)|(\d+))/
    float: /-?((\d+\.\d+[Ee][-+]?\d+)|(\d+[Ee][-+]?\d+)|(\d+\.\d+))/
    string: ESCAPED_STRING

    unknown: /[^[{;]+[^;]*/

    %import common.ESCAPED_STRING
    %import common.WS
    %ignore WS
    COMMENT: "%" /[^\n]/*
    %ignore COMMENT
"""


def arg1_construct(cls):
    return lambda self, s: cls(s[0])


class TreeToDZN(Transformer):
    @staticmethod
    def int(s):
        i = s[0]
        if i.startswith("0o") or i.startswith("-0o"):
            return int(i, 8)
        elif i.startswith("0x") or i.startswith("-0x"):
            return int(i, 16)
        else:
            return int(i)

    @staticmethod
    def item(s):
        return s[0], s[1]

    @staticmethod
    def array2d(s):
        return list(s)

    @staticmethod
    def set(s):
        if len(s) == 1:
            return set(s[0])
        else:
            return range(s[0], s[1] + 1)

    @staticmethod
    def string(s):
        return str(s[0][1:-1])

    @staticmethod
    def true(_):
        return True

    @staticmethod
    def false(_):
        return False

    items = dict
    unknown = arg1_construct(UnknownExpression)
    list = list
    array = arg1_construct(lambda i: i)
    ident = arg1_construct(str)
    float = arg1_construct(float)
    value = arg1_construct(lambda i: i)


dzn_parser = Lark(dzn_grammar, start="items", parser="lalr")


def parse_dzn(dzn: Union[Path, str]):
    if isinstance(dzn, Path):
        dzn = dzn.read_text()
    tree = dzn_parser.parse(dzn)
    dzn_dict = TreeToDZN().transform(tree)
    return dzn_dict



# type: ignore
# flake8: noqa

# DO NOT EDIT -- AUTOMATICALLY GENERATED FROM:
# https://github.com/Dekker1/iro-mzn

# EDIT: This file was edited in the end anyway because the current version of
# Iro doesn't seems to correctly generate the "pop" instructions for the
# different rule sets


import re

from pygments.lexer import RegexLexer, bygroups
from pygments.token import (
    Comment,
    Generic,
    Keyword,
    Literal,
    Name,
    Number,
    Operator,
    Punctuation,
    String,
)

__all__ = ["MiniZincLexer"]


class MiniZincLexer(RegexLexer):
    name = "MiniZinc"
    aliases = ["fzn", "dzn", "mzn", "minizinc"]
    filenames = ["*.mzn", "*.fzn", "*.dzn"]
    flags = re.MULTILINE | re.UNICODE

    tokens = {
        "root": [
            ("(/\\*)", bygroups(Comment), "multi_line_comment__1"),
            ("(%.*)", bygroups(Comment)),
            ("(@)", bygroups(Generic.Inserted), "main__1"),
            ("(\\b0o[0-7]+)", bygroups(Number)),
            ("(\\b0x[0-9A-Fa-f]+)", bygroups(Number)),
            ("(\\b0x[0-9A-Fa-f]+)", bygroups(Number)),
            ("(\\b\\d+(?:(?:.\\d+)?[Ee][-+]?\\d+|.\\d+))", bygroups(Number)),
            ("(\\b\\d+)", bygroups(Number)),
            ('(\\")', bygroups(String), "string__1"),
            ("(\\b(?:true|false)\\b)", bygroups(Literal)),
            ("(\\bnot\\b|<->|->|<-|\\\\/|\\bxor\\b|/\\\\)", bygroups(Operator)),
            ("(<|>|<=|>=|==|=|!=)", bygroups(Operator)),
            ("(\\+|-|\\*|/|\\bdiv\\b|\\bmod\\b)", bygroups(Operator)),
            (
                "(\\b(?:in|subset|superset|union|diff|symdiff|intersect|\\.\\.)\\b)",
                bygroups(Operator),
            ),
            ("(;)", bygroups(Punctuation)),
            ("(:)", bygroups(Punctuation)),
            ("(,)", bygroups(Punctuation)),
            ("(\\{)", bygroups(Punctuation), "main__2"),
            ("(\\[)", bygroups(Punctuation), "main__3"),
            ("(\\()", bygroups(Punctuation), "main__4"),
            ("(\\}|\\]|\\))", bygroups(Generic.Error)),
            ("(\\|)", bygroups(Generic.Error)),
            (
                "(\\b(?:annotation|constraint|function|include|op|output|minimize|maximize|predicate|satisfy|solve|test|type)\\b)",
                bygroups(Keyword),
            ),
            (
                "(\\b(?:ann|array|bool|enum|float|int|list|of|par|set|string|tuple|var|record|any|opt)\\b)",
                bygroups(Keyword.Type),
            ),
            (
                "(\\b(?:for|forall|exists|if|then|elseif|else|endif|where|let|in)\\b)",
                bygroups(Keyword),
            ),
            ("(\\b(?:case|op)\\b)", bygroups(Generic.Error)),
            (
                "(\\b(?:abort|abs|acosh|array_intersect|array_union|array1d|array2d|array3d|array4d|array5d|array6d|asin|assert|atan|bool2int|card|ceil|concat|cos|cosh|dom|dom_array|dom_size|fix|exp|floor|index_set|index_set_1of2|index_set_2of2|index_set_1of3|index_set_2of3|index_set_3of3|int2float|is_fixed|join|lb|lb_array|length|ln|log|log2|log10|min|max|pow|product|round|set2array|show|show_int|show_float|sin|sinh|sqrt|sum|tan|tanh|trace|ub|ub_array)\\b)",
                bygroups(Name.Builtin),
            ),
            (
                "(\\b(?:circuit|disjoint|maximum|maximum_arg|member|minimum|minimum_arg|network_flow|network_flow_cost|partition_set|range|roots|sliding_sum|subcircuit|sum_pred)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:alldifferent|all_different|all_disjoint|all_equal|alldifferent_except_0|nvalue|symmetric_all_different)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:lex2|lex_greater|lex_greatereq|lex_less|lex_lesseq|strict_lex2|value_precede|value_precede_chain)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:arg_sort|decreasing|increasing|sort)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:int_set_channel|inverse|inverse_set|link_set_to_booleans)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:among|at_least|at_most|at_most1|count|count_eq|count_geq|count_gt|count_leq|count_lt|count_neq|distribute|exactly|global_cardinality|global_cardinality_closed|global_cardinality_low_up|global_cardinality_low_up_closed)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:bin_packing|bin_packing_capa|bin_packing_load|diffn|diffn_k|diffn_nonstrict|diffn_nonstrict_k|geost|geost_bb|geost_smallest_bb|knapsack)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:alternative|cumulative|disjunctive|disjunctive_strict|span)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            ("(\\b(?:regular|regular_nfa|table)\\b)", bygroups(Name.Builtin.Pseudo)),
            (
                "(\\b[A-Za-z][A-Za-z0-9_]*|'[^'\\n\\r]*')(\\()",
                bygroups(Name.Function, Punctuation),
                "main__5",
            ),
            ("(\\b[A-Za-z][A-Za-z0-9_]*|'[^'\\n\\r]*')", bygroups(Name.Variable)),
            ("(\n|\r|\r\n)", String),
            (".", String),
        ],
        "main__1": [
            ("(@)", bygroups(Generic.Inserted), "#pop"),
            ("(\n|\r|\r\n)", String),
            (".", Generic.Inserted),
        ],
        "main__2": [
            ("(\\})", bygroups(Punctuation), "#pop"),
            ("(\\|)", bygroups(Punctuation)),
            ("(/\\*)", bygroups(Comment), "multi_line_comment__1"),
            ("(%.*)", bygroups(Comment)),
            ("(@)", bygroups(Generic.Inserted), "main__1"),
            ("(\\b0o[0-7]+)", bygroups(Number)),
            ("(\\b0x[0-9A-Fa-f]+)", bygroups(Number)),
            ("(\\b0x[0-9A-Fa-f]+)", bygroups(Number)),
            ("(\\b\\d+(?:(?:.\\d+)?[Ee][-+]?\\d+|.\\d+))", bygroups(Number)),
            ("(\\b\\d+)", bygroups(Number)),
            ('(\\")', bygroups(String), "string__1"),
            ("(\\b(?:true|false)\\b)", bygroups(Literal)),
            ("(\\bnot\\b|<->|->|<-|\\\\/|\\bxor\\b|/\\\\)", bygroups(Operator)),
            ("(<|>|<=|>=|==|=|!=)", bygroups(Operator)),
            ("(\\+|-|\\*|/|\\bdiv\\b|\\bmod\\b)", bygroups(Operator)),
            (
                "(\\b(?:in|subset|superset|union|diff|symdiff|intersect|\\.\\.)\\b)",
                bygroups(Operator),
            ),
            ("(;)", bygroups(Punctuation)),
            ("(:)", bygroups(Punctuation)),
            ("(,)", bygroups(Punctuation)),
            ("(\\{)", bygroups(Punctuation), "#push"),
            ("(\\[)", bygroups(Punctuation), "main__3"),
            ("(\\()", bygroups(Punctuation), "main__4"),
            ("(\\}|\\]|\\))", bygroups(Generic.Error)),
            ("(\\|)", bygroups(Generic.Error)),
            (
                "(\\b(?:annotation|constraint|function|include|op|output|minimize|maximize|predicate|satisfy|solve|test|type)\\b)",
                bygroups(Keyword),
            ),
            (
                "(\\b(?:ann|array|bool|enum|float|int|list|of|par|set|string|tuple|var|record|any|opt)\\b)",
                bygroups(Keyword.Type),
            ),
            (
                "(\\b(?:for|forall|exists|if|then|elseif|else|endif|where|let|in)\\b)",
                bygroups(Keyword),
            ),
            ("(\\b(?:case|op)\\b)", bygroups(Generic.Error)),
            (
                "(\\b(?:abort|abs|acosh|array_intersect|array_union|array1d|array2d|array3d|array4d|array5d|array6d|asin|assert|atan|bool2int|card|ceil|concat|cos|cosh|dom|dom_array|dom_size|fix|exp|floor|index_set|index_set_1of2|index_set_2of2|index_set_1of3|index_set_2of3|index_set_3of3|int2float|is_fixed|join|lb|lb_array|length|ln|log|log2|log10|min|max|pow|product|round|set2array|show|show_int|show_float|sin|sinh|sqrt|sum|tan|tanh|trace|ub|ub_array)\\b)",
                bygroups(Name.Builtin),
            ),
            (
                "(\\b(?:circuit|disjoint|maximum|maximum_arg|member|minimum|minimum_arg|network_flow|network_flow_cost|partition_set|range|roots|sliding_sum|subcircuit|sum_pred)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:alldifferent|all_different|all_disjoint|all_equal|alldifferent_except_0|nvalue|symmetric_all_different)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:lex2|lex_greater|lex_greatereq|lex_less|lex_lesseq|strict_lex2|value_precede|value_precede_chain)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:arg_sort|decreasing|increasing|sort)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:int_set_channel|inverse|inverse_set|link_set_to_booleans)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:among|at_least|at_most|at_most1|count|count_eq|count_geq|count_gt|count_leq|count_lt|count_neq|distribute|exactly|global_cardinality|global_cardinality_closed|global_cardinality_low_up|global_cardinality_low_up_closed)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:bin_packing|bin_packing_capa|bin_packing_load|diffn|diffn_k|diffn_nonstrict|diffn_nonstrict_k|geost|geost_bb|geost_smallest_bb|knapsack)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:alternative|cumulative|disjunctive|disjunctive_strict|span)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            ("(\\b(?:regular|regular_nfa|table)\\b)", bygroups(Name.Builtin.Pseudo)),
            (
                "(\\b[A-Za-z][A-Za-z0-9_]*|'[^'\\n\\r]*')(\\()",
                bygroups(Name.Function, Punctuation),
                "main__5",
            ),
            ("(\\b[A-Za-z][A-Za-z0-9_]*|'[^'\\n\\r]*')", bygroups(Name.Variable)),
            ("(\n|\r|\r\n)", String),
            (".", String),
        ],
        "main__3": [
            ("(\\])", bygroups(Punctuation), "#pop"),
            ("(\\|)", bygroups(Punctuation)),
            ("(/\\*)", bygroups(Comment), "multi_line_comment__1"),
            ("(%.*)", bygroups(Comment)),
            ("(@)", bygroups(Generic.Inserted), "main__1"),
            ("(\\b0o[0-7]+)", bygroups(Number)),
            ("(\\b0x[0-9A-Fa-f]+)", bygroups(Number)),
            ("(\\b0x[0-9A-Fa-f]+)", bygroups(Number)),
            ("(\\b\\d+(?:(?:.\\d+)?[Ee][-+]?\\d+|.\\d+))", bygroups(Number)),
            ("(\\b\\d+)", bygroups(Number)),
            ('(\\")', bygroups(String), "string__1"),
            ("(\\b(?:true|false)\\b)", bygroups(Literal)),
            ("(\\bnot\\b|<->|->|<-|\\\\/|\\bxor\\b|/\\\\)", bygroups(Operator)),
            ("(<|>|<=|>=|==|=|!=)", bygroups(Operator)),
            ("(\\+|-|\\*|/|\\bdiv\\b|\\bmod\\b)", bygroups(Operator)),
            (
                "(\\b(?:in|subset|superset|union|diff|symdiff|intersect|\\.\\.)\\b)",
                bygroups(Operator),
            ),
            ("(;)", bygroups(Punctuation)),
            ("(:)", bygroups(Punctuation)),
            ("(,)", bygroups(Punctuation)),
            ("(\\{)", bygroups(Punctuation), "main__2"),
            ("(\\[)", bygroups(Punctuation), "#push"),
            ("(\\()", bygroups(Punctuation), "main__4"),
            ("(\\}|\\]|\\))", bygroups(Generic.Error)),
            ("(\\|)", bygroups(Generic.Error)),
            (
                "(\\b(?:annotation|constraint|function|include|op|output|minimize|maximize|predicate|satisfy|solve|test|type)\\b)",
                bygroups(Keyword),
            ),
            (
                "(\\b(?:ann|array|bool|enum|float|int|list|of|par|set|string|tuple|var|record|any|opt)\\b)",
                bygroups(Keyword.Type),
            ),
            (
                "(\\b(?:for|forall|exists|if|then|elseif|else|endif|where|let|in)\\b)",
                bygroups(Keyword),
            ),
            ("(\\b(?:case|op)\\b)", bygroups(Generic.Error)),
            (
                "(\\b(?:abort|abs|acosh|array_intersect|array_union|array1d|array2d|array3d|array4d|array5d|array6d|asin|assert|atan|bool2int|card|ceil|concat|cos|cosh|dom|dom_array|dom_size|fix|exp|floor|index_set|index_set_1of2|index_set_2of2|index_set_1of3|index_set_2of3|index_set_3of3|int2float|is_fixed|join|lb|lb_array|length|ln|log|log2|log10|min|max|pow|product|round|set2array|show|show_int|show_float|sin|sinh|sqrt|sum|tan|tanh|trace|ub|ub_array)\\b)",
                bygroups(Name.Builtin),
            ),
            (
                "(\\b(?:circuit|disjoint|maximum|maximum_arg|member|minimum|minimum_arg|network_flow|network_flow_cost|partition_set|range|roots|sliding_sum|subcircuit|sum_pred)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:alldifferent|all_different|all_disjoint|all_equal|alldifferent_except_0|nvalue|symmetric_all_different)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:lex2|lex_greater|lex_greatereq|lex_less|lex_lesseq|strict_lex2|value_precede|value_precede_chain)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:arg_sort|decreasing|increasing|sort)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:int_set_channel|inverse|inverse_set|link_set_to_booleans)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:among|at_least|at_most|at_most1|count|count_eq|count_geq|count_gt|count_leq|count_lt|count_neq|distribute|exactly|global_cardinality|global_cardinality_closed|global_cardinality_low_up|global_cardinality_low_up_closed)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:bin_packing|bin_packing_capa|bin_packing_load|diffn|diffn_k|diffn_nonstrict|diffn_nonstrict_k|geost|geost_bb|geost_smallest_bb|knapsack)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:alternative|cumulative|disjunctive|disjunctive_strict|span)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            ("(\\b(?:regular|regular_nfa|table)\\b)", bygroups(Name.Builtin.Pseudo)),
            (
                "(\\b[A-Za-z][A-Za-z0-9_]*|'[^'\\n\\r]*')(\\()",
                bygroups(Name.Function, Punctuation),
                "main__5",
            ),
            ("(\\b[A-Za-z][A-Za-z0-9_]*|'[^'\\n\\r]*')", bygroups(Name.Variable)),
            ("(\n|\r|\r\n)", String),
            (".", String),
        ],
        "main__4": [
            ("(\\))", bygroups(Punctuation), "#pop"),
            ("(/\\*)", bygroups(Comment), "multi_line_comment__1"),
            ("(%.*)", bygroups(Comment)),
            ("(@)", bygroups(Generic.Inserted), "main__1"),
            ("(\\b0o[0-7]+)", bygroups(Number)),
            ("(\\b0x[0-9A-Fa-f]+)", bygroups(Number)),
            ("(\\b0x[0-9A-Fa-f]+)", bygroups(Number)),
            ("(\\b\\d+(?:(?:.\\d+)?[Ee][-+]?\\d+|.\\d+))", bygroups(Number)),
            ("(\\b\\d+)", bygroups(Number)),
            ('(\\")', bygroups(String), "string__1"),
            ("(\\b(?:true|false)\\b)", bygroups(Literal)),
            ("(\\bnot\\b|<->|->|<-|\\\\/|\\bxor\\b|/\\\\)", bygroups(Operator)),
            ("(<|>|<=|>=|==|=|!=)", bygroups(Operator)),
            ("(\\+|-|\\*|/|\\bdiv\\b|\\bmod\\b)", bygroups(Operator)),
            (
                "(\\b(?:in|subset|superset|union|diff|symdiff|intersect|\\.\\.)\\b)",
                bygroups(Operator),
            ),
            ("(;)", bygroups(Punctuation)),
            ("(:)", bygroups(Punctuation)),
            ("(,)", bygroups(Punctuation)),
            ("(\\{)", bygroups(Punctuation), "main__2"),
            ("(\\[)", bygroups(Punctuation), "main__3"),
            ("(\\()", bygroups(Punctuation), "#push"),
            ("(\\}|\\]|\\))", bygroups(Generic.Error)),
            ("(\\|)", bygroups(Generic.Error)),
            (
                "(\\b(?:annotation|constraint|function|include|op|output|minimize|maximize|predicate|satisfy|solve|test|type)\\b)",
                bygroups(Keyword),
            ),
            (
                "(\\b(?:ann|array|bool|enum|float|int|list|of|par|set|string|tuple|var|record|any|opt)\\b)",
                bygroups(Keyword.Type),
            ),
            (
                "(\\b(?:for|forall|exists|if|then|elseif|else|endif|where|let|in)\\b)",
                bygroups(Keyword),
            ),
            ("(\\b(?:case|op)\\b)", bygroups(Generic.Error)),
            (
                "(\\b(?:abort|abs|acosh|array_intersect|array_union|array1d|array2d|array3d|array4d|array5d|array6d|asin|assert|atan|bool2int|card|ceil|concat|cos|cosh|dom|dom_array|dom_size|fix|exp|floor|index_set|index_set_1of2|index_set_2of2|index_set_1of3|index_set_2of3|index_set_3of3|int2float|is_fixed|join|lb|lb_array|length|ln|log|log2|log10|min|max|pow|product|round|set2array|show|show_int|show_float|sin|sinh|sqrt|sum|tan|tanh|trace|ub|ub_array)\\b)",
                bygroups(Name.Builtin),
            ),
            (
                "(\\b(?:circuit|disjoint|maximum|maximum_arg|member|minimum|minimum_arg|network_flow|network_flow_cost|partition_set|range|roots|sliding_sum|subcircuit|sum_pred)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:alldifferent|all_different|all_disjoint|all_equal|alldifferent_except_0|nvalue|symmetric_all_different)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:lex2|lex_greater|lex_greatereq|lex_less|lex_lesseq|strict_lex2|value_precede|value_precede_chain)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:arg_sort|decreasing|increasing|sort)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:int_set_channel|inverse|inverse_set|link_set_to_booleans)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:among|at_least|at_most|at_most1|count|count_eq|count_geq|count_gt|count_leq|count_lt|count_neq|distribute|exactly|global_cardinality|global_cardinality_closed|global_cardinality_low_up|global_cardinality_low_up_closed)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:bin_packing|bin_packing_capa|bin_packing_load|diffn|diffn_k|diffn_nonstrict|diffn_nonstrict_k|geost|geost_bb|geost_smallest_bb|knapsack)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:alternative|cumulative|disjunctive|disjunctive_strict|span)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            ("(\\b(?:regular|regular_nfa|table)\\b)", bygroups(Name.Builtin.Pseudo)),
            (
                "(\\b[A-Za-z][A-Za-z0-9_]*|'[^'\\n\\r]*')(\\()",
                bygroups(Name.Function, Punctuation),
                "main__5",
            ),
            ("(\\b[A-Za-z][A-Za-z0-9_]*|'[^'\\n\\r]*')", bygroups(Name.Variable)),
            ("(\n|\r|\r\n)", String),
            (".", String),
        ],
        "main__5": [
            ("(\\))", bygroups(Punctuation), "#pop"),
            ("(/\\*)", bygroups(Comment), "multi_line_comment__1"),
            ("(%.*)", bygroups(Comment)),
            ("(@)", bygroups(Generic.Inserted), "main__1"),
            ("(\\b0o[0-7]+)", bygroups(Number)),
            ("(\\b0x[0-9A-Fa-f]+)", bygroups(Number)),
            ("(\\b0x[0-9A-Fa-f]+)", bygroups(Number)),
            ("(\\b\\d+(?:(?:.\\d+)?[Ee][-+]?\\d+|.\\d+))", bygroups(Number)),
            ("(\\b\\d+)", bygroups(Number)),
            ('(\\")', bygroups(String), "string__1"),
            ("(\\b(?:true|false)\\b)", bygroups(Literal)),
            ("(\\bnot\\b|<->|->|<-|\\\\/|\\bxor\\b|/\\\\)", bygroups(Operator)),
            ("(<|>|<=|>=|==|=|!=)", bygroups(Operator)),
            ("(\\+|-|\\*|/|\\bdiv\\b|\\bmod\\b)", bygroups(Operator)),
            (
                "(\\b(?:in|subset|superset|union|diff|symdiff|intersect|\\.\\.)\\b)",
                bygroups(Operator),
            ),
            ("(;)", bygroups(Punctuation)),
            ("(:)", bygroups(Punctuation)),
            ("(,)", bygroups(Punctuation)),
            ("(\\{)", bygroups(Punctuation), "main__2"),
            ("(\\[)", bygroups(Punctuation), "main__3"),
            ("(\\()", bygroups(Punctuation), "main__4"),
            ("(\\}|\\]|\\))", bygroups(Generic.Error)),
            ("(\\|)", bygroups(Generic.Error)),
            (
                "(\\b(?:annotation|constraint|function|include|op|output|minimize|maximize|predicate|satisfy|solve|test|type)\\b)",
                bygroups(Keyword),
            ),
            (
                "(\\b(?:ann|array|bool|enum|float|int|list|of|par|set|string|tuple|var|record|any|opt)\\b)",
                bygroups(Keyword.Type),
            ),
            (
                "(\\b(?:for|forall|exists|if|then|elseif|else|endif|where|let|in)\\b)",
                bygroups(Keyword),
            ),
            ("(\\b(?:case|op)\\b)", bygroups(Generic.Error)),
            (
                "(\\b(?:abort|abs|acosh|array_intersect|array_union|array1d|array2d|array3d|array4d|array5d|array6d|asin|assert|atan|bool2int|card|ceil|concat|cos|cosh|dom|dom_array|dom_size|fix|exp|floor|index_set|index_set_1of2|index_set_2of2|index_set_1of3|index_set_2of3|index_set_3of3|int2float|is_fixed|join|lb|lb_array|length|ln|log|log2|log10|min|max|pow|product|round|set2array|show|show_int|show_float|sin|sinh|sqrt|sum|tan|tanh|trace|ub|ub_array)\\b)",
                bygroups(Name.Builtin),
            ),
            (
                "(\\b(?:circuit|disjoint|maximum|maximum_arg|member|minimum|minimum_arg|network_flow|network_flow_cost|partition_set|range|roots|sliding_sum|subcircuit|sum_pred)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:alldifferent|all_different|all_disjoint|all_equal|alldifferent_except_0|nvalue|symmetric_all_different)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:lex2|lex_greater|lex_greatereq|lex_less|lex_lesseq|strict_lex2|value_precede|value_precede_chain)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:arg_sort|decreasing|increasing|sort)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:int_set_channel|inverse|inverse_set|link_set_to_booleans)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:among|at_least|at_most|at_most1|count|count_eq|count_geq|count_gt|count_leq|count_lt|count_neq|distribute|exactly|global_cardinality|global_cardinality_closed|global_cardinality_low_up|global_cardinality_low_up_closed)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:bin_packing|bin_packing_capa|bin_packing_load|diffn|diffn_k|diffn_nonstrict|diffn_nonstrict_k|geost|geost_bb|geost_smallest_bb|knapsack)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:alternative|cumulative|disjunctive|disjunctive_strict|span)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            ("(\\b(?:regular|regular_nfa|table)\\b)", bygroups(Name.Builtin.Pseudo)),
            (
                "(\\b[A-Za-z][A-Za-z0-9_]*|'[^'\\n\\r]*')(\\()",
                bygroups(Name.Function, Punctuation),
                "#push",
            ),
            ("(\\b[A-Za-z][A-Za-z0-9_]*|'[^'\\n\\r]*')", bygroups(Name.Variable)),
            (".", String),
        ],
        "multi_line_comment__1": [
            ("(\\*/)", bygroups(Comment), "#pop"),
            ("(\n|\r|\r\n)", String),
            (".", Comment),
        ],
        "string__1": [
            ('(\\")', bygroups(String), "#pop"),
            ("(\\\\\\()", bygroups(Punctuation), "string__2"),
            ("(\\\\[\"'\\\\nrvt])", bygroups(String.Escape)),
            ('([^\\"\\\\\\n\\r]+)', bygroups(String)),
            ("(\n|\r|\r\n)", String),
            (".", String),
        ],
        "string__2": [
            ("(\\))", bygroups(Punctuation), "#pop"),
            ("(/\\*)", bygroups(Comment), "multi_line_comment__1"),
            ("(%.*)", bygroups(Comment)),
            ("(@)", bygroups(Generic.Inserted), "main__1"),
            ("(\\b0o[0-7]+)", bygroups(Number)),
            ("(\\b0x[0-9A-Fa-f]+)", bygroups(Number)),
            ("(\\b0x[0-9A-Fa-f]+)", bygroups(Number)),
            ("(\\b\\d+(?:(?:.\\d+)?[Ee][-+]?\\d+|.\\d+))", bygroups(Number)),
            ("(\\b\\d+)", bygroups(Number)),
            ('(\\")', bygroups(String), "string__1"),
            ("(\\b(?:true|false)\\b)", bygroups(Literal)),
            ("(\\bnot\\b|<->|->|<-|\\\\/|\\bxor\\b|/\\\\)", bygroups(Operator)),
            ("(<|>|<=|>=|==|=|!=)", bygroups(Operator)),
            ("(\\+|-|\\*|/|\\bdiv\\b|\\bmod\\b)", bygroups(Operator)),
            (
                "(\\b(?:in|subset|superset|union|diff|symdiff|intersect|\\.\\.)\\b)",
                bygroups(Operator),
            ),
            ("(;)", bygroups(Punctuation)),
            ("(:)", bygroups(Punctuation)),
            ("(,)", bygroups(Punctuation)),
            ("(\\{)", bygroups(Punctuation), "main__2"),
            ("(\\[)", bygroups(Punctuation), "main__3"),
            ("(\\()", bygroups(Punctuation), "main__4"),
            ("(\\}|\\]|\\))", bygroups(Generic.Error)),
            ("(\\|)", bygroups(Generic.Error)),
            (
                "(\\b(?:annotation|constraint|function|include|op|output|minimize|maximize|predicate|satisfy|solve|test|type)\\b)",
                bygroups(Keyword),
            ),
            (
                "(\\b(?:ann|array|bool|enum|float|int|list|of|par|set|string|tuple|var|record|any|opt)\\b)",
                bygroups(Keyword.Type),
            ),
            (
                "(\\b(?:for|forall|exists|if|then|elseif|else|endif|where|let|in)\\b)",
                bygroups(Keyword),
            ),
            ("(\\b(?:case|op)\\b)", bygroups(Generic.Error)),
            (
                "(\\b(?:abort|abs|acosh|array_intersect|array_union|array1d|array2d|array3d|array4d|array5d|array6d|asin|assert|atan|bool2int|card|ceil|concat|cos|cosh|dom|dom_array|dom_size|fix|exp|floor|index_set|index_set_1of2|index_set_2of2|index_set_1of3|index_set_2of3|index_set_3of3|int2float|is_fixed|join|lb|lb_array|length|ln|log|log2|log10|min|max|pow|product|round|set2array|show|show_int|show_float|sin|sinh|sqrt|sum|tan|tanh|trace|ub|ub_array)\\b)",
                bygroups(Name.Builtin),
            ),
            (
                "(\\b(?:circuit|disjoint|maximum|maximum_arg|member|minimum|minimum_arg|network_flow|network_flow_cost|partition_set|range|roots|sliding_sum|subcircuit|sum_pred)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:alldifferent|all_different|all_disjoint|all_equal|alldifferent_except_0|nvalue|symmetric_all_different)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:lex2|lex_greater|lex_greatereq|lex_less|lex_lesseq|strict_lex2|value_precede|value_precede_chain)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:arg_sort|decreasing|increasing|sort)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:int_set_channel|inverse|inverse_set|link_set_to_booleans)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:among|at_least|at_most|at_most1|count|count_eq|count_geq|count_gt|count_leq|count_lt|count_neq|distribute|exactly|global_cardinality|global_cardinality_closed|global_cardinality_low_up|global_cardinality_low_up_closed)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:bin_packing|bin_packing_capa|bin_packing_load|diffn|diffn_k|diffn_nonstrict|diffn_nonstrict_k|geost|geost_bb|geost_smallest_bb|knapsack)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            (
                "(\\b(?:alternative|cumulative|disjunctive|disjunctive_strict|span)\\b)",
                bygroups(Name.Builtin.Pseudo),
            ),
            ("(\\b(?:regular|regular_nfa|table)\\b)", bygroups(Name.Builtin.Pseudo)),
            (
                "(\\b[A-Za-z][A-Za-z0-9_]*|'[^'\\n\\r]*')(\\()",
                bygroups(Name.Function, Punctuation),
                "main__5",
            ),
            ("(\\b[A-Za-z][A-Za-z0-9_]*|'[^'\\n\\r]*')", bygroups(Name.Variable)),
            ("(\n|\r|\r\n)", String),
            (".", String),
        ],
    }
